% ---------------------------------------------------------------------------
% -- This MATLAB file was developed by Daniel Llamocca.  It may be
% -- freely copied and/or distributed at no cost.  Any persons using this
% -- file for any purpose do so at their own risk, and are responsible for
% -- the results of such use.  Daniel Llamocca does not guarantee that
% -- this file is complete, correct, or fit for any particular purpose.
% -- NO WARRANTY OF ANY KIND IS EXPRESSED OR IMPLIED.  This notice must
% -- accompany any copy of this file.
% -------------------------------------------------------------------------
% -

function x = gen_input(NX, nstreams, B, tb, mode)
% mode  : Specifies that inputs are generated tailored to either the dynamic
%         or the static system. mode can be 'stc' or 'dyn'
% NX    : number of input sample values
% nstreams: number of streams to be created and concatenated (Static case)
%                   dynamic case: no effect            
% tb    : 'yes' -> create input text file for Modelsim testbench 
%                  create input binary file for hardware testing
%         Dynamic case: Only 1 stream is created for binary file
%         Static case: A concatenated stream is created for testbench and binary file
% B = input values' bitwidth, B <= 32

q = quantizer ('fixed','round','saturate',[B B-1]);
qt = quantizer ('ufixed','round','saturate',[B 0]);

% Input Generation:
  t = linspace(-pi/2 + 0.01, 5*pi/2 - 0.01, NX);
  figure;
  for i = 1:nstreams
      func{i} = sin(t + pi/i)*(rem(i,2) == 1) + cos(t + pi/i)*(rem(i,2) == 0);
      x(i,:) = quantize (q, func{i});
      hold on; plot(x(i,:));
  end
% 'x' is a matrix of 'nstreams' rows and 'NX' columns

% *****************************************
%               clear x;
%               IM = rgb2gray(imread('im1.jpg')); % 640x480 image
%               [n m] = size (IM); % n: # of rows, m: # of columns
%               
%             % Modifying pixel range. IM belongs to [0,255]. We want it to belong to
%             % [-127,128], and ultimately to [-1,1) with 8 bits in 2's complement
%             % This is because the FPGA HW assumes the input to belong to [-1,1)
%             % Recall to use imshow(image,[]) when showing the images.
%               IC = double(IM)-128; % -128 <= IC <= 127
%               ICf = IC/128; % -1 <= ICf < 1                            
%               x = ICf;
% % *****************************************
  
  x_h = num2hex(q,x'); % We use transpose 'cause the order of x_h follows
                       % the columns of x (first column, then next column)
                       
if mode == 'stc'
    % Generating input text file for the Modelsim testbench
      strm = [];
      for i = 1:nstreams
          strm = [strm x(i,:)];
      end
   
      strm_h = num2hex(q, strm);
      filename = strcat('in_bench_sin_',num2str(B),'_',num2str(nstreams),'_',num2str(NX),'.txt');
      fod = fopen (filename, 'wt');
      for k = 1:length(strm)
          fprintf (fod, '%s\n', strm_h(k,:));
      end
      fclose (fod);
      
    % Generating input binary file for hardware testing
      % Keep in mind that we are converting the hexadecimal values to integer
      % values, in order to create the input binary file. The hardware will get
      % these values as bits.
      % The choice of unsigned integer is just to make the process clear. A
      % signed integer could have also been chosen. Actually, the
      % definition of 'qt' is necessary since fwrite does not allow us to
      % write fixed-point values with fractional part, only integer part.
      
      for i = 1:nstreams*NX
          strm_f(i) = hex2num(qt,strm_h(i,:));
      end
      
      % strm_f contain the 
      fod = fopen ('tst_stc.bif', 'w');
      if B <= 8
          fwrite(fod, strm_f, 'uint8'); % strm_f is a row vector
      elseif B <= 16
          fwrite(fod, strm_f, 'uint16'); % strm_f is a row vector
      elseif B <= 32
          fwrite(fod, strm_f, 'uint32'); % strm_f is a row vector
      else
          error ('B can not be larger than 32!');
      end
      
      fclose(fod);
     
elseif mode == 'dyn'
  
     % Generating binary file
      % Keep in mind that we are converting the hexadecimal values to integer
      % values, in order to create the input binary file. The hardware will get
      % these values as bits.
      for i = 1:nstreams*NX
          xf(i) = hex2num(qt,x_h(i,:));
          % this gets 'x' in a raster scan fashion
      end

      fod = fopen ('tst_dyn.bif', 'w');
      if B <= 8
          fwrite(fod, xf, 'uint8'); % xf is a row vector
      elseif B <= 16
          fwrite(fod, xf, 'uint16'); % xf is a row vector
      elseif B <= 32
          fwrite(fod, xf, 'uint32'); % xf is a row vector
      else
          error ('B can not be larger than 32!');
      end

      fclose (fod);

      % Reading binary file.. FOR TESTING!! it works!!!
%          fid = fopen ('tst_mat.bif', 'r');
%          if B <= 8
%             c = fread(fid,nstreams*NX, 'uint8 => uint8');
%          elseif B <= 16
%             c = fread(fid,nstreams*NX, 'uint16 => uint16');
%          elseif B <= 32
%             c = fread(fid,nstreams*NX, 'uint32 => uint32');
%          else
%             error( 'error');
%          end if
%          fclose(fid);
         % c is column vector
end
