% ---------------------------------------------------------------------------
% -- This MATLAB file was developed by Daniel Llamocca.  It may be
% -- freely copied and/or distributed at no cost.  Any persons using this
% -- file for any purpose do so at their own risk, and are responsible for
% -- the results of such use.  Daniel Llamocca does not guarantee that
% -- this file is complete, correct, or fit for any particular purpose.
% -- NO WARRANTY OF ANY KIND IS EXPRESSED OR IMPLIED.  This notice must
% -- accompany any copy of this file.
% -------------------------------------------------------------------------
% -

function [] = gen_txt_LUT(NH,L,N, h, symmetry, filename_header)
% GEN_TXT_LUT (NH, L, N, h, symmetry, filename_header) creates one or
% several 'txt' files that contain the LUT values to be used by the VHDL
% filter architecture.
%
%  h: Set of filters, h{k}: kth filter
%  NH: bitwidth of filter coefficients
%  L: # of inputs of LUTs.
%  N:  Number of coefficients filter h[n]
% symmetry = 'y', 'a': same case as far the coefficients concern
%              These 2 cases result in the exact same LUT file
% symmetry = 'n' --> no symmetry, LUT file is different
% Outputs:
%   textfile = 'LUT_values.txt' if the # of filters in h is 1 and
%                 'filename_header' is not used.
%              filename_header.txt if the # of filters in h is 1 and
%                 'filename_header' is indeed used.
%   textfile = 'LUT_values[i].txt', i = 1:n if the # of filters in h is 'n'
%               and the parameter 'filename_header' is not used
%   textfile = filename_header{rows,col} if the # of filters in h is 2
%               and the parameter 'filename_header' is used
%
% Assumptions:
% - Coefficients format: [NH NH-1]

  LO = NH + ceil(log2(L)); % number of bits of the LUT output
  % Quantizer for the L-input LUT outputs:
    qb = quantizer ('fixed', 'round', 'saturate', [LO NH-1]); % the # of fractional bits remain the same
    NF = size(h,2); % Number of filters

    % Quantizer for the L-input LUT outputs
    qL = quantizer ('ufixed', 'round', [L 0]);
    if NF == 1
        if nargin == 6 % if 'filename_header is used'
            textfile = strcat(filename_header,'.txt');
        else
            textfile = 'LUT_values.txt';
        end
    end

    if symmetry == 'y' || symmetry == 'a'
        M = ceil(N/2);
    else
        M = N;
    end
    
    if rem(M,L) ~= 0
        error ('# of effective taps (M) is not multiple of L');
    end
    
    for k = 1:NF % There are NF Filters        
        if NF ~= 1
            if nargin == 6
                if k == 1
                    textfile = strcat(filename_header,'rows','.txt');
                elseif k == 2                    
                    textfile = strcat(filename_header,'cols','.txt');
                else
                    error('In this case the # of filters can only be 2');
                end
            else
                textfile = ['LUT_values[',num2str(k),'].txt'];
            end
        end
        
        fid(k) = fopen (textfile,'wt'); % generates text file in write mode                 
        
        for i = 1:M/L % There are M/L block filters
%             fb(1) = 0;
%             fb(2) = h{k}((i-1)*NL+1); % h(1)
%             fb(3) = h{k}((i-1)*NL+2); ;% h(2)
%             fb(4) = h{k}((i-1)*NL+1) + h{k}((i-1)*NL+2); % h(1) + h(2)
%             fb(5) = h{k}((i-1)*NL+3); % h(3)
%             fb(6) = h{k}((i-1)*NL+3) + h{k}((i-1)*NL+1); % h(3) + h(1)
%             fb(7) = h{k}((i-1)*NL+3) + h{k}((i-1)*NL+2); % h(3) + h(2)
%             fb(8) = h{k}((i-1)*NL+3) + h{k}((i-1)*NL+2) + h{k}((i-1)*NL+1); % h(3) + h(2) + h(1)
%             fb(9) = h{k}(i*NL); % h(4)
%             fb(10) = h{k}(i*NL) + h{k}((i-1)*NL+1); % h(4) + h(1)
%             fb(11) = h{k}(i*NL) + h{k}((i-1)*NL+2); % h(4) + h(2)
%             fb(12) = h{k}(i*NL) + h{k}((i-1)*NL+2) + h{k}((i-1)*NL+1); % h(4) + h(2) + h(1)
%             fb(13) = h{k}(i*NL) + h{k}((i-1)*NL+3); % h(4) + h(3)
%             fb(14) = h{k}(i*NL) + h{k}((i-1)*NL+3) + h{k}((i-1)*NL+1); % h(4) + h(3) + h(1)
%             fb(15) = h{k}(i*NL) + h{k}((i-1)*NL+3) + h{k}((i-1)*NL+2); % h(4) + h(3) + h(2)
%             fb(16) = h{k}(i*NL) + h{k}((i-1)*NL+3) + h{k}((i-1)*NL+2) + h{k}((i-1)*NL+1); % h(4) + h(3) + h(2) + h(1)
            for j = 1: 2^L
                bits = num2bin(qL,j-1); % |bits(1)|bits(2)|bits(3)|... |bits(L)| --> bits1 = MSB, bits(L) = LSB
                fb(j) = 0;
                % h(1) = h((i-1)*L+ 1) is multiplied by bits(L)
                % h(2) = h((i-1)*L+ 2) is multiplied by bits(L-1)
                % .......
                % h(L) = h((i-1)*L+ L) is multiplied by bits(1)
                for m = 1:L
                    %fb(j) = fb(j) + h{k}((i-1)*L+m)*bin2num(qL,bits(L-m+1));
                    fb(j) = fb(j) + h{k}((i-1)*L+m)*(bits(L-m+1)=='1');
                end
            end
            % Now we check if fb(16) is the maximum negative number we can
            % get with the representation [LO NH-1].
            % If it is, we change it by adding 1 bit to the LSB, so that
            % fb(16) is not the maximum negative number any more

            if fb(2^L) == -2^(LO-NH) % maximum negative number with [LO NH-1]
                fb(2^L) = fb(2^L) + 2^(-(NH-1)); % 2^(-(NH-1)): minimum positive number with [LO NH-1]
            end

            fB = -fb; % Negative set of LUT values
            fB_bin = num2bin(qb, fB);
            fb_bin = num2bin(qb, fb);

            % For a particular block filter, we create fb and its negative,
            % fB. Each filter block requires both fb and fB since in each
            % filter block one LUT is the one that outputs fB.
            for j = 1:2^L
                fprintf (fid(k), '%s\n', fb_bin(j,:));
            end
            
            fprintf (fid(k),'L\n');
            
            for j = 1:2^L
                fprintf (fid(k), '%s\n', fB_bin(j,:));
            end
            
            if i ~= M/L
                fprintf (fid(k),'L\n');
            end
            
        end

        fclose (fid(k));
    end
